@@ -148,7 +148,7 @@ unpack_internal (gfc_array_char *ret, const gfc_array_char *vector,
              the next dimension.  */
           count[n] = 0;
           /* We could precalculate these products, but this is a less
-             frequently used path so proabably not worth it.  */
+             frequently used path so probably not worth it.  */
           rptr -= rstride[n] * extent[n];
           fptr -= fstride[n] * extent[n];
           mptr -= mstride[n] * extent[n];
